import re
from sage.ext.sage_object import SageObject

class MPolynomialParser(SageObject):
  """
  This class parses a string representation of a multivariate polynomial
  over a finite extension field and returns a matching MPolynomial.

  It does the same job as eval() but:
  1. it is not as feature complete, it only accepts a very narrow
     set of polynomials and base fields
  2. it is slower for small polynomials
  3. it is much _faster_ for large polynomials and repetetive conversions

  The speed is achieved through:
  1. hashing of finite field elements
  2. by using the internal polynomial representation
  3. no polynomial multiplications, exponentations and additions are
     used
  """

  def __init__(self,ring):
    """
    constructs a new string parser for the given ring.
    """
    self.ring = ring
    self.a = self.ring.base_ring().gen()
    self.a_repr = str(self.a)
    self.x_repr = str(self.ring.gen(0))[0]
    self.c_dict = {}

    self.exp_st = [ str(i) for i in self.ring.gens() ]
    self.exp_st.reverse()

  def parse_string(self,st,i=int(0)):
    """
    Parses a given string starting at i.

    The resulting dictionary will be placed in self.polydict

    """
    self.exps = [int(0),]*self.ring.ngens()
    self.polydict = {}
    self.c = int(0)
    self._parse_string(st,i)

  def _parse_string(self,st,i=int(0)):
    """
    """
    while i < len(st):
      if st[i]=="(":
        i=int(self._parse_string(st,int(i+1)))
      elif st[i]==self.x_repr:
        if self.c==0:
          self.c=int(1)
        i=int(self._parse_monomial(st,int(i)))
      elif st[i]==self.a_repr or st[i]=="0" or st[i]=="1": #regexp here
        i=int(self._parse_coef(st,int(i)))
      elif st[i]=="*":
        i+=int(1)
      elif st[i]==" ":
        i+=int(1)
      elif st[i]=="+":
        self.exps.reverse()
        self.polydict[tuple(self.exps)]=self.c
        self.c=int(0)
        self.exps = [int(0),]*self.ring.ngens()
        i+=int(1)
      elif st[i]==")":
        return int(i+1)
      else:
        raise "ParseError","unknown literal '%s' around '%s' with i=%d"%(st[i],st[i-5:i+5],i)
	
    self.exps.reverse()
    if self.c!=0:
      self.polydict[tuple(self.exps)]=self.c

    return i

  def _parse_coef(self,st,i):
    """
    Parses a coefficient in the base_ring()
    """
    start = i
    i=int(self.__parse_coef(st,int(i)))
    c = st[start:i-1]#.replace("^","**")
    if not self.c_dict.has_key(c):
      self.c_dict[c]=eval(c,{},{"a":self.a})
    self.c = self.c_dict[c]
    return int(i-1)

  def __parse_coef(self,st,end):
    """
    
    """
    if st[end]=="(":
      end=int(self.__parse_coef(st,int(end+1)))
    elif st[end]==")":
      return int(end+1)
    else:
      s=st[end:].find("(")
      e=st[end:].find(")")
      if -1 < s < e :
        end+=s
      elif -1 < e:
        end+=e
      else:
        return len(st)+1
    return self.__parse_coef(st,end)

  def _parse_monomial(self,st,i):
    """
    parses a single monomial
    """
    for j in range(len(self.exp_st)):
      if st[i:].startswith(self.exp_st[j]):
         #found variable
         l = len(self.exp_st[j])
         if st[i+l:].startswith("**"):
           mtch=re.match("[0-9]+",st[i+l+2:])
           self.exps[j]=int(st[i+l+2:i+l+2+mtch.end()])
           return int(i+l+2+mtch.end())
         else:
           self.exps[j]=int(1)
           return int(i+l)
      else:
        continue
      return 0

  def mpolynomial(self,st=None):
    """
    Returns a MPolynomial matching the string st
    """
    ret = self.ring(0)
    if st != None:
      self.parse_string(st)
    ret._MPolynomial__element._PolyDict__repn=self.polydict
    return ret

  def _repr_(self):
    return "MPolynomialParser for %s"%(self.ring)

def singular_multiply(left,right,mpp=None):
  """
  Multiplies a Matrix/Vector with a Matrix/Vector in Singular.

  It is assumed that 'left' and 'right' are over a MPolynomial
  over a finite extension field.

  It is assumed that this function is faster for large matrices
  than the native left*right implementation.
  

  INPUT:
      left  -- Matrix over MPolynomialRing
      right -- Matrix over MPolynomialRing
      mpp   -- MPolynomialParser for MPolynomialRing

  OUTPUT:
      a Matrix over MPolynomialRing == left*right

  WARNING:
      It is assumed that a correct singular ring is set.

  EXAMPLES:
      sage: r = MPolynomialRing(GF(2**8),2,'x')
      sage: ms=MatrixSpace(r,10,10)
      sage: left=ms.random_element()
      sage: right=ms.random_element()
      sage: left*right==singular_multiply(left,right)
      _ = True
  """

  ring = right.base_ring()
  
  #convert to singular
  left=singular.matrix(left.nrows(),left.ncols(),left.list())
  right=singular.matrix(right.nrows(),right.ncols(),right.list())

  #do some work
  _product = left*right

  #convert to sage
  product = Matrix(ring, _product.nrows(), _product.ncols())

  if mpp == None:
    mpp = MPolynomialParser(ring)

  for y in range(product.nrows()):
    for x in range(product.ncols()):
      product[y,x] = mpp.mpolynomial( _product[y+1,x+1].sage_polystring() )

  return product


def singular_map(dictionary, gens, singular_ring):
    """
    Returns a singular map for the given dictionary. This dictionary
    must have variables as keys.

    INPUT:
        dictionary    -- a dict() with ring variables as keys and base_ring values
                         as values.
        gens          -- MPolynomialRing generators
        singular_ring -- singular ring to work in

    OUTPUT:
        SingularObject representing the singular map

    EXAMPLES:
        sage: r=MPolynomialRing(GF(2**8),10,'x')
        sage: s_r = singular_ring(r)
        sage: d = {r.gen(0):r.base_ring().gen(0)}
        sage: singular_map(d,r.gens(),s_r)
        _ =
        sage11[1]=(a)
        sage11[2]=x_1
        sage11[3]=x_2
        sage11[4]=x_3
        sage11[5]=x_4
        sage11[6]=x_5
        sage11[7]=x_6
        sage11[8]=x_7
        sage11[9]=x_8
        sage11[10]=x_9
     """

    gens = list(gens) #local copy
    
    for i in range(0,len(gens)):
        if gens[i] in dictionary:
            gens[i] = dictionary[gens[i]]

    map_name = singular._next_var_name()
    map_str = "map %s = %s"%(map_name,singular_ring.name())
    for elem in gens:
        map_str +=",%s"%(elem)
    singular.eval(map_str)
    return singular(map_name)

def fix_poly(self, mapping, parser):
    """
    fixes a polynomial using singular and a given singular mapping.

    INPUT:
        self    -- MPolynomial to fix
        mapping -- Singular mapping 
        parser  -- MPolynomialParser over self's ring

    OUTPUT:
        fixed MPolynomial

    EXAMPLES:
        sage: r=MPolynomialRing(GF(2**8),10,'x')
        sage: s_r = singular_ring(r)
        sage: d = {r.gen(0):r.base_ring().gen(0)}
        sage: m = singular_map(d,r.gens(),s_r)
        sage: r.gen(0)**5+r.gen(3)*r.gen(4)
        _ = x_3*x_4 + x_0^5
        sage: f=r.gen(0)**5+r.gen(3)*r.gen(4)
        sage: fix_poly(f,m,mpp)
        _ = a^5 + x_3*x_4

    """
    if self.is_constant():
      return self

    f=singular(self._singular_init_())
    ret = mapping(f)
    return parser.mpolynomial(ret.sage_polystring())
    

def singular_ring(ring):
    """
    Returns a singular ring for a given MPolynomial
    over a finite extension field.

    INPUT:
       ring -- MPolynomial over finite extension ring


    OUTPUT:
       singular ring matching 'ring'

    EXAMPLES:
       sage: r=MPolynomialRing(GF(2**8),10,'x')
       sage: singular_ring(r)
       //   characteristic : 2
       //   1 parameter    : a
       //   minpoly        : (a^8+a^4+a^3+a^2+1)
       //   number of vars : 10
       //        block   1 : ordering lp
       //                  : names    x_0 x_1 x_2 x_3 x_4 x_5 x_6 x_7 x_8 x_9
       //        block   2 : ordering C

    WARNING:
        Rings over finite extension fields are implemented
        in Singular using a global 'minpoly=<modulus>' call.
        This minpoly is not stored in this ring. So if another
        ring is set e.g. using set_ring() and this ring is
        set again using set_ring() the minpoly will be lost and
        the ring will be useless.
    """

    gen = str(ring.base_ring().gen())

    if ring.ngens()==1:
      _vars = str(ring.gen())
    else:
      _vars = str(ring.gens())
    
    r = singular.ring( "(%s,%s)"%(ring.characteristic(),gen),
                       _vars,
                       ring.term_order().singular_str() )
    minpoly = "minpoly=%s;"%(str(ring.base_ring().modulus()).replace("x",gen))
    singular.eval(minpoly)
    verbose("new singular ring: %s"%(r.name()), level=2)
    return r


