#! /usr/bin/env python
#@+leo-ver=4
#@+node:@file pyrexembed.py
#@@first

# written by David McNab (david@freenet.org.nz)
# copyright (C) 2005 by David McNab, all rights reserved
# released under the GNU Lesser General Public License

"""
pyrexembed.py

Utility to extract embedded C++ code elements from specially
marked-up Pyrex files, collate them and generate:
    - correctly-formatted pyrex file
    - C++ source file, containing the embedded code
    - C header file, which is included into the generated
      pyrex file

Marked-up pyrex files must have a .pyxe extension. A file called
fred.pyxe will be processed with the outputs:
    - fred.pyx
    - fred_embed.cpp
    - fred_embed.h

There are only two kinds of markup - header and embed.

Header markup does not need to an embedded function,
and will typically contain #include statements, #defines etc. Everything
inside a header markup will be copied literally to the generated .cpp file,
but not the generated .h file.

A header markup section commences with a line starting with
'#header{', and ends with a line starting with '#}header', for example::

    #header{
    #include <someheader.h>
    #define ENABLE_SOME_FEATURE 1
    #}header

Embedding markup allows for convenient declaration of 'shim' functions
which can be called from within Pyrex code, since they will be written to
the generated .cpp file as 'extern "C"'.

Embedded markup takes the form::

    #embed{    <c function declaration>
        <body of c function, containing C++ code if desired>
    #}embed

The C function declaration will be extracted and written to:
    - the generated pyrex .pyx file, within a
      'cdef extern from <myfile_embed.h>:' block
    - the generated .h file, verbatim
    - the generated .cpp file, along with the function body,
      with an 'extern "C"' specifier

For example::

    cdef class MyExtension:

        cdef void *thisptr

        def __init__(self, size):

            #embed{    void *Fred_new(int size)
                Fred *obj = new Fred(size);
                return (void *)obj;
            #}embed

            self.thisptr = Fred_new(size)

        def onemethod(self, arg1):
            "Some method of my pyrex extension class"

            #embed{    int Fred_onemethod(void *fredobj, long arg1)
               int result = ((Fred *)fredobj)->onemethod(arg1);
                   return result;
            #}embed

            # try to convert arg to int
            arg1int = int(arg1)
            
            # now run the shim func
            return Fred_onemethod(self.thisptr, arg1int)

"""

#@+others
#@+node:imports
import sys, os, re

#@-node:imports
#@+node:globals
argv = sys.argv
argc = len(argv)
progname = argv[0]

#@-node:globals
#@+node:class PyrexEmbedder
class PyrexEmbedder:
    """
    Processes .pyxe files (Pyrex files with C++ code inside markups)
    into valid Pyrex (.pyx) files, plus accompanying .cpp and .h files

    Refer to module docstring for usage information
    """
    #@    @+others
    #@+node:process
    def process(self, pathname, filetype=".cpp"):
        """
        Perform the processing. Argument must be an absolute or
        relative pathname of the input file, which must exist and have
        a .pyxe extension.
    
        For further information, refer class and module docstrings.
    
        Arguments:
            - pathname - path of .pyxe input file
            - filetype - type of generated shim file, default '.cpp'
        """
        # barf if no such file
        if not os.path.isfile(pathname):
            raise Exception("No such file: '%s'" % pathname)
    
        # break up pathname into component parts
        filebase, ext = os.path.splitext(pathname)
        filename = os.path.split(pathname)[1]
    
        # barf if not a .pyxe file
        if ext != ".pyxe":
            raise Exception("File '%s' should have a .pyxe extension" % pathname)
    
        # derive and save file names
        self.pathIn = pathname
        self.pathPyrex = filebase + ".pyx"
        self.pathCpp = filebase + "_embed" + filetype
        self.pathH = filebase + "_embed.h"
        self.pathInFile = filename + ".pyxe"
    
        # open input and track the line number
        self.fileIn = file(pathname)
        self.lineNum = 0
    
        # set up working data structures
        self.pyrexLines = []
        self.pyrexHeaderLines = []
        self.cppHeaderLines = []
        self.embedFuncs = []
    
        # process, initially as Pyrex code
        self.processPyrex()
    
        # now generate the output
        self.generateFiles()
    
    #@-node:process
    #@+node:processPyrex
    def processPyrex(self):
        """
        Reads and processes pure pyrex code from input
        """
        while True:
            # get next line from .pyxi
            line = self.readline()
    
            # bail if we're finished processing
            if line == None:
                break
    
            # break it up
            parts = re.split("\\s+", line.strip(), 1)
            firstbit = parts[0]
    
            # detect markups
            if firstbit == "#header{":
                # grab header stuff
                self.processHeader()
            elif firstbit == "#embed{":
                # barf if no C funcdec
                if len(parts) < 2:
                    self.error("'#embed{' declaration must have C function declarator")
                # ok, get declarator
                decl = parts[1] # fixme - this mandates that the whole dec fit on one line
                self.processEmbed(decl)
            elif firstbit == '#}embed':
                self.error("Not in a '#embed{' block")
            elif firstbit == '#}header':
                self.error("Not in a '#header{' block")
            else:
                # normal pyrex source line - save it
                self.pyrexLines.append(line)
    #@-node:processPyrex
    #@+node:processHeader
    def processHeader(self):
        """
        Reads and process header code from input
    
        Expect this to end with a '#}header'
        """
        while True:
            # get next line
            line = self.readline()
    
            # complain at eof
            if line == None:
                self.error("Unexpected end of file during '#header{' block")
    
            # break it up
            firstbit = re.split("\\s+", line.strip().lower(), 1)[0]
            if firstbit == '#}header':
                # done
                break
            elif firstbit == '#header{':
                self.error("Illegally nested #header{")
            elif firstbit in ['#embed{', '#}embed']:
                self.error("Illegal markup within header block")
            else:
                # legal - add to header lines
                self.cppHeaderLines.append(line)
    #@-node:processHeader
    #@+node:processEmbed
    def processEmbed(self, decl):
        """
        Reads and processes embed code from input
        """
        startline = self.lineNum
    
        embLines = []
        while True:
            # get next line
            line = self.readline()
            
            # complain if EOF
            if line == None:
                self.error(
                    "Unexpected end of file during '#embed{' block (%s)" % startline)
    
            # break it up
            firstbit = re.split("\\s+", line.strip().lower(), 1)[0]
            if firstbit == '#}embed':
                # done
                break
            elif firstbit == '#embed{':
                self.error("Illegally nested #embed{")
            elif firstbit in ['#header{', '#}header']:
                self.error("Illegal markup within embed block")
            else:
                # legal - add to embed lines
                embLines.append(line)
    
        # ok, got full embedded func def
        self.embedFuncs.append((decl.strip(), "\n".join(embLines) + "\n"))
    #@-node:processEmbed
    #@+node:generateFiles
    def generateFiles(self):
        """
        Generate the required .pyx, .cpp and .h files from what we've
        extracted from the input .pyxe file
        """
        #self.pyrexLines = []
        #self.pyrexHeaderLines = []
        #self.cppHeaderLines = []
        #self.embedFuncs = []
    
        # open all the output files
        filePyx = file(self.pathPyrex, "w")
        fileCpp = file(self.pathCpp, "w")
        fileH = file(self.pathH, "w")
    
        # save shim import funcs
        pyrexShimDecs = []
    
        # write in an initial comment into each file
        commentLines = [
            "**************************************************************",
            "Generated by pyrexembed from %s" % self.pathInFile,
            "Please do not edit",
            "**************************************************************",
            ]
        for line in commentLines:
            filePyx.write("# %s\n" % line)
            fileCpp.write("// %s\n" % line)
            fileH.write("// %s\n" % line)
        for f in (filePyx, fileCpp, fileH):
            f.write("\n")
    
        # write out all the header stuff into the cpp file
        fileCpp.write("// verbatim header lines from #header{ markups\n")
        for line in self.cppHeaderLines:
            fileCpp.write(line + "\n")
        fileCpp.write("// end of verbatim header lines\n\n")
    
        # now process all the shim functions
        for dec, body in self.embedFuncs:
            # add the pyrex import
            pyrexShimDecs.append("    cdef "+dec)
            
            # add the .h dec
            fileH.write("extern " + dec + ";\n")
            
            # add the full dec and body to .c file
            fileCpp.write("extern \"C\" " + dec + "\n{\n" + body + "}\n\n")
    
        # if there are shim funcs, stick in a pyrex dec block
        if pyrexShimDecs:
            filePyx.write('cdef extern from "%s":\n' % os.path.split(self.pathH)[1])
            filePyx.write("\n".join(pyrexShimDecs) + "\n\n")
    
        # now write out the rest of the pyrex stuff
        filePyx.write("\n".join(self.pyrexLines) + "\n")
    
        # should be all done now
        filePyx.close()
        fileCpp.close()
        fileH.close()
    
    #@-node:generateFiles
    #@+node:readline
    def readline(self):
        """
        Reads a line of input file, updating the line number
        """
        if not self.fileIn:
            raise Exception("Internal error - reading from closed input file")
    
        line = self.fileIn.readline()
        if line == '':
            # need None as a sentinel, since we're removing
            # trailing newlines
            self.fileIn.close()
            self.fileIn = None
            return None
        line = line[:-1]    # chomp off trailing newline
        self.lineNum += 1
        return line
    #@-node:readline
    #@+node:error
    def error(self, msg):
        raise Exception("%s:\n  %s: %s:\s  %s" % (
            progname,
            self.pathInFile,
            self.lineNum,
            msg)
            )
    
    
    #@-node:error
    #@-others

#@-node:class PyrexEmbedder
#@+node:process
def process(pathname, filetype=".cpp"):
    """
    Arguments:
        - pathname - path of .pyxe input file
        - filetype - type of generated shim file, default '.cpp'
    """

    PyrexEmbedder().process(pathname, filetype)

#@-node:process
#@+node:linkwithgplusplus
def linkwithgplusplus():
    """
    changes the linker from gcc to g++
    """
    from distutils import sysconfig

    # grab original init func
    save_init_posix = sysconfig._init_posix

    # create a new one that uses old one, but changes from gcc to c++
    def my_init_posix():
        print 'my_init_posix: changing gcc to g++'
        save_init_posix()
        g = sysconfig._config_vars
        g['LINKCC'] = 'g++ -pthread'
        g['LDSHARED'] = 'g++ -shared'

    # and plug in the new function
    sysconfig._init_posix = my_init_posix

#@-node:linkwithgplusplus
#@+node:main
def main():
    
    if argc not in [2, 3]:
        print "usage: %s filename.pyxi [filetype]"
        print "filetype defaults to .cpp"
        sys.exit(1)
    
    path = argv[1]
    if argc == 2:
        process(path)
    else:
        process(path, argv[2])



#@-node:main
#@+node:mainline
if __name__ == '__main__':
    main()

#@-node:mainline
#@-others

#@-node:@file pyrexembed.py
#@-leo
